/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */

package zswzg.kex.tls;

import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeSet;

import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.Data.NumberArray;
import de.bsvrz.dav.daf.main.Data.ReferenceArray;
import de.bsvrz.dav.daf.main.Data.ReferenceValue;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Klasse für die grundlegende (default) Umsetzung der Anzeige-Informationen.
 * Umsetzung KExTls nach AnzeigenGlobal und Umsetzung AnzeigenGlobal nach KExTls.
 *
 * @version $Revision: 1.4 $ / $Date: 2012/05/23 15:47:43 $ / ($Author: Drapp $)
 *
 * @author Dambach-Werke GmbH
 * @author Thomas Drapp
 *
 */
public class AnzeigeTypUmsetzungTlsDefault
{
  /**
   * DebugLogger für Debug-Ausgaben.
   */
  private static Debug m_debug = Debug.getLogger();
  /**
   * Status ok.
   */
  final static int STATUS_OK = 0;
  /**
   * Status TeilStörung.
   */
  final static int STATUS_TEILSTOERUNG = 1;
  /**
   * Status gestört.
   */
  final static int STATUS_GESTOERT = 2;
  /**
   * Status Stromausfall.
   */
  final static int STATUS_STROMAUSFALL = 3;
  /**
   * Status Türkontakt.
   */
  final static int STATUS_TUERKONTAKT = 4;
  /**
   * Status undefiniert.
   */
  final static int STATUS_UNDEFINIERT = 5;
  /**
   * Status Kommunikationsausfall.
   */
  final static int STATUS_KOMMUNIKATIONSAUSFALL = 6;
  /**
   * Kanalsteuerung wird von atg.tlsGloDeFehler übernommen, daher muss atg.tlsGloKanalSteuerung nicht ausgewertet werden
   */
  final static boolean m_kanalSteuerungUeberDeFehler = true;
  /**
   * HashMap mit AnzeigeAktorTls mit Key DEs.
   */
  protected HashMap m_listeAnzeigeAktorTlsDes = null;
  /**
   * HashMap mit AnzeigeAktorTls mit Key Anzeige.
   */
  protected HashMap m_listeAnzeigeAktorTlsAnzeige = null;
  /**
   * HashMap mit Attributgruppe atg.anzeigeEigenschaftIst für alle Anzeigen aktuell.
   */
  protected HashMap m_listeAnzeigeEigenschaftIst = null;
  /**
   * HashMap für Status Methoden für alle Anzeigen.
   */
  protected HashMap m_listeAnzeigeStatus = null;
  /**
   * Information, ob die Anzeige freie Grafik darstellen kann.
   */
  protected boolean m_kannFreieGrafik;
  /**
   * Information, ob die Anzeige freien Text darstellen kann.
   */
  protected boolean m_kannFreiText;

  /**
   * Konstruktor mit Übergabe der notwendigen HashMaps.
   *
   * @param listeAnzeigeAktorTlsDes Liste mit AnzeigeAktorTls mit Key DEs.
   * @param listeAnzeigeAktorTlsAnzeige Liste mit AnzeigeQuerschnittAktorTls mit Key Anzeige.
   * @param listeAnzeigeEigenschaftIst Liste mit Attributgruppe atg.anzeigeEigenschaftIst für alle Anzeigen aktuell.
   * @param listeAnzeigeStatus Liste mit Status Methoden für alle Anzeigen.
   */
  public AnzeigeTypUmsetzungTlsDefault(HashMap listeAnzeigeAktorTlsDes, HashMap listeAnzeigeAktorTlsAnzeige, HashMap listeAnzeigeEigenschaftIst, HashMap listeAnzeigeStatus)
  {
    m_listeAnzeigeAktorTlsDes = listeAnzeigeAktorTlsDes;
    m_listeAnzeigeAktorTlsAnzeige = listeAnzeigeAktorTlsAnzeige;
    m_listeAnzeigeEigenschaftIst = listeAnzeigeEigenschaftIst;
    m_listeAnzeigeStatus = listeAnzeigeStatus;

    m_kannFreieGrafik = false;
    m_kannFreiText = false;

    m_debug.finer("AnzeigeTypUmsetzungTlsDefault");
  }

  /**
   * Konstruktor mit Übergabe der notwendigen HashMaps und speziellen Anzeige-Informationen.
   *
   * @param listeAnzeigeAktorTlsDes Liste mit AnzeigeAktorTls mit Key DEs.
   * @param listeAnzeigeAktorTlsAnzeige Liste mit AnzeigeQuerschnittAktorTls mit Key Anzeige.
   * @param kannFreieGrafik <code>true</code>, wenn Anzeige freie Grafik darstellen kann, sonst <code>false</code>.
   * @param kannFreiText <code>true</code>, wenn Anzeige freien Text darstellen kann, sonst <code>false</code>.
   */
  public AnzeigeTypUmsetzungTlsDefault(HashMap listeAnzeigeAktorTlsDes, HashMap listeAnzeigeAktorTlsAnzeige, boolean kannFreieGrafik, boolean kannFreiText)
  {
    m_listeAnzeigeAktorTlsDes = listeAnzeigeAktorTlsDes;
    m_listeAnzeigeAktorTlsAnzeige = listeAnzeigeAktorTlsAnzeige;
    m_listeAnzeigeEigenschaftIst = null;
    m_listeAnzeigeStatus = null;

    m_kannFreieGrafik = kannFreieGrafik;
    m_kannFreiText = kannFreiText;

    m_debug.finer("AnzeigeTypUmsetzungTlsDefault");
  }

  /**
   * Abarbeitung des DE-Fehler genau nach TLS.
   *
   * @param result Attributgruppe atg.tlsGloDeFehler.
   * @return Data
   */
  public Data tlsGloDeFehler(ResultData result)
  {
    SystemObject object = result.getObject();

    AnzeigeAktorTls anzeigeAktor = (AnzeigeAktorTls)getListeAnzeigeAktorTlsDes().get(object.getPid());

    if (anzeigeAktor == null)
      return null;

    String anzeigePid = anzeigeAktor.getAnzeige().getPid();

    Data dataAnzeigeEigenschaftIst = (Data)getListeAnzeigeEigenschaftIst().get(anzeigePid);

    m_debug.finer("tlsGloDeFehler DE " + object.getPid() + "\nAnzeige " + anzeigePid + "\nresult " + result + "\nData " + dataAnzeigeEigenschaftIst);

    if (result.isNoSourceAvailable()) // keine Quelle vorhanden
    {
      ZwischenschichtAq.atgAnzeigeEigenschaftIstInitialisieren(dataAnzeigeEigenschaftIst, result.getDataTime(), statusBestimmen(anzeigePid, "tlsGloDeFehler", STATUS_KOMMUNIKATIONSAUSFALL));
    }
    else // Quelle vorhanden
    {
      if (result.hasData()) // Daten vorhanden
      {
        Data data = result.getData();

        String deFehlerStatus = data.getTextValue("DEFehlerStatus").getValueText();

        if (deFehlerStatus.equals("ok"))
        {
          dataAnzeigeEigenschaftIst.getScaledValue("Status").setText(statusBestimmen(anzeigePid, "tlsGloDeFehler", STATUS_OK));
//          dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getTimeValue("Zeit").setMillis(result.getDataTime());
        }
        else
        {
          ZwischenschichtAq.atgAnzeigeEigenschaftIstInitialisieren(dataAnzeigeEigenschaftIst, result.getDataTime(), statusBestimmen(anzeigePid, "tlsGloDeFehler", STATUS_GESTOERT));
        }

        String deKanalStatus = data.getTextValue("DEKanalStatus").getValueText();
        String passiviert = deKanalStatus.equals("aktiv") ? "Nein" : "Ja";
        dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getScaledValue("Passiviert").setText(passiviert);
      }
      else // keine Daten vorhanden
      {
        ZwischenschichtAq.atgAnzeigeEigenschaftIstInitialisieren(dataAnzeigeEigenschaftIst, result.getDataTime(), statusBestimmen(anzeigePid, "tlsGloDeFehler", STATUS_UNDEFINIERT));
      }
    }

    return dataAnzeigeEigenschaftIst;
  }

  /**
   * Abarbeitung der Kanalsteuerung genau nach TLS.
   * Wird über DE-Fehler abgearbeitet, deshalb ausser Betrieb.
   *
   * @param result Attributgruppe atg.tlsGloKanalSteuerung.
   * @return Data
   */
  public Data tlsGloKanalSteuerung(ResultData result)
  {
    if (m_kanalSteuerungUeberDeFehler)
      return null;

    SystemObject object = result.getObject();

    AnzeigeAktorTls anzeigeAktor = (AnzeigeAktorTls)getListeAnzeigeAktorTlsDes().get(object.getPid());

    if (anzeigeAktor == null)
      return null;

    String anzeigePid = anzeigeAktor.getAnzeige().getPid();

    Data dataAnzeigeEigenschaftIst = (Data)getListeAnzeigeEigenschaftIst().get(anzeigePid);

    if (dataAnzeigeEigenschaftIst == null)
      return null;

    m_debug.finer("tlsGloKanalSteuerung DE " + object.getPid() + "\nAnzeige " + anzeigePid + "\nresult " + result + "\nData " + dataAnzeigeEigenschaftIst);

    if (result.isSourceAvailable()) // Quelle vorhanden
    {
      if (result.hasData()) // Daten vorhanden
      {
        Data data = result.getData();

        String deKanalStatus = data.getTextValue("DEKanalStatus").getValueText();
        String passiviert = deKanalStatus.equals("aktiv") ? "Nein" : "Ja";
        dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getScaledValue("Passiviert").setText(passiviert);
      }
      else // keine Daten vorhanden
      {
        dataAnzeigeEigenschaftIst = null; // keine Änderung melden
      }
    }

    return dataAnzeigeEigenschaftIst;
  }

  /**
   * Abarbeitung der gestörten Textpositionen genau nach TLS.
   *
   * @param result Attributgruppe atg.tlsWzgGestörteTextPositionen.
   * @return Data
   */
  public Data tlsWzgGestörteTextPositionen(ResultData result)
  {
    SystemObject object = result.getObject();

    AnzeigeAktorTls anzeigeAktor = (AnzeigeAktorTls)getListeAnzeigeAktorTlsDes().get(object.getPid());

    if (anzeigeAktor == null)
      return null;

    String anzeigePid = anzeigeAktor.getAnzeige().getPid();

    Data dataAnzeigeEigenschaftIst = (Data)getListeAnzeigeEigenschaftIst().get(anzeigePid);

    m_debug.finer("tlsWzgGestörteTextPositionen DE " + object.getPid() + "\nAnzeige " + anzeigePid + "\nresult " + result + "\nData " + dataAnzeigeEigenschaftIst);

    if (result.isNoSourceAvailable()) // keine Quelle vorhanden
    {
//      ZwischenschichtAq.atgAnzeigeEigenschaftIstInitialisieren(dataAnzeigeEigenschaftIst, result.getDataTime(), statusBestimmen(anzeigePid, "tlsWzgGestörteTextPositionen", STATUS_KOMMUNIKATIONSAUSFALL));
    }
    else // Quelle vorhanden
    {
      if (result.hasData()) // Daten vorhanden
      {
        Data data = result.getData();

//        dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getTimeValue("Zeit").setMillis(result.getDataTime());

        int[] gestörteTextPositionen = data.getUnscaledArray("GestörteTextPosition").getIntArray();

        NumberArray array = dataAnzeigeEigenschaftIst.getUnscaledArray("GestörteTextpositionen");

        array.setLength(gestörteTextPositionen.length);

        for (int i = 0; i < gestörteTextPositionen.length; i++)
        {
          array.getValue(i).set(gestörteTextPositionen[i]);
        }

        int status = STATUS_OK;

        if (gestörteTextPositionen.length > 0)
        {
          status = STATUS_TEILSTOERUNG;
        }

        dataAnzeigeEigenschaftIst.getScaledValue("Status").setText(statusBestimmen(anzeigePid, "tlsWzgGestörteTextPositionen", status));
      }
      else // keine Daten vorhanden
      {
        dataAnzeigeEigenschaftIst = null; // keine Änderung melden
      }
    }

    return dataAnzeigeEigenschaftIst;
  }

  /**
   * Abarbeitung der nicht darstellbaren WVZ genau nach TLS 93.
   *
   * @param result Attributgruppe atg.tlsWzgNichtDarstellbareWVZ.
   * @return Data
   */
  public Data tlsWzgNichtDarstellbareWVZ(ResultData result)
  {
    SystemObject object = result.getObject();

    AnzeigeAktorTls anzeigeAktor = (AnzeigeAktorTls)getListeAnzeigeAktorTlsDes().get(object.getPid());

    if (anzeigeAktor == null)
      return null;

    String anzeigePid = anzeigeAktor.getAnzeige().getPid();

    Data dataAnzeigeEigenschaftIst = (Data)getListeAnzeigeEigenschaftIst().get(anzeigePid);

    m_debug.finer("tlsWzgNichtDarstellbareWVZ DE " + object.getPid() + "\nAnzeige " + anzeigePid + "\nresult " + result + "\nData " + dataAnzeigeEigenschaftIst);

    if (result.isNoSourceAvailable()) // keine Quelle vorhanden
    {
//      ZwischenschichtAq.atgAnzeigeEigenschaftIstInitialisieren(dataAnzeigeEigenschaftIst, result.getDataTime(), statusBestimmen(anzeigePid, "tlsWzgNichtDarstellbareWVZ", STATUS_KOMMUNIKATIONSAUSFALL));
    }
    else // Quelle vorhanden
    {
      if (result.hasData()) // Daten vorhanden
      {
        Data data = result.getData();

        AnzeigeTypUmsetzungTls anzeigeTypUmsetzungTls = anzeigeAktor.getAnzeigeTypUmsetzung();

//        dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getTimeValue("Zeit").setMillis(result.getDataTime());

        int[] nichtDarstellbareWVZ = data.getUnscaledArray("NichtDarstellbareWVZ").getIntArray();

        ReferenceArray array = dataAnzeigeEigenschaftIst.getReferenceArray("GestörteInhalte");

        array.setLength(nichtDarstellbareWVZ.length);

        HashMap mapWvzInhalt = null;

        mapWvzInhalt = anzeigeTypUmsetzungTls.getListeWvzInhaltUmsetzungTlsCode();

        for (int i = 0; i < nichtDarstellbareWVZ.length; i++)
        {
          WvzInhaltUmsetzungTls wvzInhaltUmsetzungTls = (WvzInhaltUmsetzungTls)mapWvzInhalt.get(nichtDarstellbareWVZ[i]);

          if (wvzInhaltUmsetzungTls != null)
          {
            array.getReferenceValue(i).setText(wvzInhaltUmsetzungTls.getPidWvzInhalt());
          }
          else
          {
            array.getReferenceValue(i).setText("null");
          }
        }

        int status = STATUS_OK;

        if (nichtDarstellbareWVZ.length > 0)
        {
          status = STATUS_TEILSTOERUNG;
        }

        dataAnzeigeEigenschaftIst.getScaledValue("Status").setText(statusBestimmen(anzeigePid, "tlsWzgNichtDarstellbareWVZ", status));
      }
      else // keine Daten vorhanden
      {
        dataAnzeigeEigenschaftIst = null; // keine Änderung melden
      }
    }

    return dataAnzeigeEigenschaftIst;
  }

  /**
   * Abarbeitung des TLS-WechselText genau nach TLS 93.
   *
   * D.h. in WVZTextZeichen sind nur sichtbare Zeichen enthalten, keine Steuerzeichen für erweiterte Funktionalitäten.
   *
   * @param result Attributgruppe atg.tlsWzgWechselText.
   * @return Data
   */
  public Data tlsWzgWechselText(ResultData result)
  {
    SystemObject object = result.getObject();

    AnzeigeAktorTls anzeigeAktor = (AnzeigeAktorTls)getListeAnzeigeAktorTlsDes().get(object.getPid());

    if (anzeigeAktor == null)
      return null;

    String anzeigePid = anzeigeAktor.getAnzeige().getPid();

    Data dataAnzeigeEigenschaftIst = (Data)getListeAnzeigeEigenschaftIst().get(anzeigePid);

    m_debug.finer("tlsWzgWechselText DE " + object.getPid() + "\nAnzeige " + anzeigePid + "\nresult " + result + "\nData " + dataAnzeigeEigenschaftIst);

    if (result.isNoSourceAvailable()) // keine Quelle vorhanden
    {
      ZwischenschichtAq.atgAnzeigeEigenschaftIstInitialisieren(dataAnzeigeEigenschaftIst, result.getDataTime(), "Kommunikationsausfall");
    }
    else // Quelle vorhanden
    {
      if (result.hasData()) // Daten vorhanden
      {
        Data data = result.getData();

        AnzeigeTypUmsetzungTls anzeigeTypUmsetzungTls = anzeigeAktor.getAnzeigeTypUmsetzung();

//      Ergebnis löschen

        dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getItem("DynamischerWechseltext").getReferenceValue("Font").setSystemObject(null);
        dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getItem("DynamischerWechseltext").getUnscaledValue("TextAusrichtung").set(-1);
        dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getItem("DynamischerWechseltext").getReferenceArray("Textdefinition").setLength(0);

//      Font bestimmen

        String pidFontUmsetzungTls = anzeigeTypUmsetzungTls.getPidGrundFont();

        String zeichenfolge = data.getTextValue("WVZTextZeichen").getValueText();
        int laengeInput = zeichenfolge.length();
        int laengeText = laengeInput;

        if (laengeInput == 0) // Anzeige ausgeschaltet
        {
          // nichts weiter zu tun
        }
        else
        {
          FontUmsetzungTls fontUmsetzungTls = null;

          if (pidFontUmsetzungTls != null)
          {
            HashMap mapFont = anzeigeTypUmsetzungTls.getListeFontUmsetzungTls();

            if (mapFont != null)
              fontUmsetzungTls = (FontUmsetzungTls)mapFont.get(pidFontUmsetzungTls);
          }

          HashMap mapZeichen = null;

          if (fontUmsetzungTls != null)
            mapZeichen = fontUmsetzungTls.getListeZeichenUmsetzungTlsCode();

          if ((mapZeichen == null) || (mapZeichen.size() == 0))
            mapZeichen = anzeigeTypUmsetzungTls.getListeZeichenUmsetzungTlsCode();

//        Pid Font über Umsetzung bestimmen

          String pidFont = null;

          if (fontUmsetzungTls != null)
          {
            pidFont = fontUmsetzungTls.getPidFont();
          }

          if (pidFont != null)
          {
            dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getItem("DynamischerWechseltext").getReferenceValue("Font").setText(pidFont);
          }

//        Text

          dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getItem("DynamischerWechseltext").getReferenceArray("Textdefinition").setLength(laengeText);

          for (int i = 0; i < zeichenfolge.length(); i++)
          {
            m_debug.finest("Zeichen " + i + " " + zeichenfolge.codePointAt(i));

            int code = zeichenfolge.codePointAt(i);

            ZeichenUmsetzungTls zeichenUmsetzungTls = (ZeichenUmsetzungTls)mapZeichen.get(code);

            if (zeichenUmsetzungTls != null)
              dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getItem("DynamischerWechseltext").getReferenceArray("Textdefinition").getReferenceValue(i).setText(zeichenUmsetzungTls.getPidZeichen());
            else
              dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getItem("DynamischerWechseltext").getReferenceArray("Textdefinition").getReferenceValue(i).setSystemObject(null);
          }
        }

        tlsWzgWechselTextBlinken(data, dataAnzeigeEigenschaftIst);

        dataAnzeigeEigenschaftIst.getScaledValue("Status").setText(statusBestimmen(anzeigePid, "tlsWzgWechselText", STATUS_OK));

        tlsWzgWechselTextBlinken(data, dataAnzeigeEigenschaftIst);
      }
      else // keine Daten vorhanden
      {
        tlsWzgWechselTextBlinken(null, dataAnzeigeEigenschaftIst);

        dataAnzeigeEigenschaftIst.getScaledValue("Status").setText(statusBestimmen(anzeigePid, "tlsWzgWechselText", STATUS_UNDEFINIERT));
      }
    }

    return dataAnzeigeEigenschaftIst;
  }

  /**
   * Abarbeitung des TLS-WvzStellZustand nach TLS 93.
   *
   * @param result Attributgruppe atg.tlsWzgWvzStellZustand.
   * @return Data
   */
  public Data tlsWzgWvzStellZustand(ResultData result)
  {
    SystemObject object = result.getObject();

    AnzeigeAktorTls anzeigeAktor = (AnzeigeAktorTls)getListeAnzeigeAktorTlsDes().get(object.getPid());

    if (anzeigeAktor == null)
      return null;

    String anzeigePid = anzeigeAktor.getAnzeige().getPid();

    Data dataAnzeigeEigenschaftIst = (Data)getListeAnzeigeEigenschaftIst().get(anzeigePid);

    m_debug.finer("tlsWzgWvzStellZustand DE " + object.getPid() + "\nAnzeige " + anzeigePid + "\nresult " + result + "\nData " + dataAnzeigeEigenschaftIst);

    if (result.isNoSourceAvailable()) // keine Quelle vorhanden
    {
      ZwischenschichtAq.atgAnzeigeEigenschaftIstInitialisieren(dataAnzeigeEigenschaftIst, result.getDataTime(), statusBestimmen(anzeigePid, "tlsWzgWvzStellZustand", STATUS_KOMMUNIKATIONSAUSFALL));
    }
    else // Quelle vorhanden
    {
//      dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getTimeValue("Zeit").setMillis(result.getDataTime());
      dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getReferenceValue("AnzeigeInhalt").setSystemObject(null);

      if (result.hasData()) // Daten vorhanden
      {
        Data data = result.getData();

        AnzeigeTypUmsetzungTls anzeigeTypUmsetzungTls = anzeigeAktor.getAnzeigeTypUmsetzung();

        HashMap mapWvzInhalt = null;

        mapWvzInhalt = anzeigeTypUmsetzungTls.getListeWvzInhaltUmsetzungTlsCode();

        int code = data.getScaledValue("WvzCode").intValue();

        m_debug.finest("Code " + code);

        WvzInhaltUmsetzungTls wvzInhaltUmsetzungTls = (WvzInhaltUmsetzungTls)mapWvzInhalt.get(code);

        if (wvzInhaltUmsetzungTls != null)
        {
          dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getReferenceValue("AnzeigeInhalt").setText(wvzInhaltUmsetzungTls.getPidWvzInhalt());
        }

        tlsWzgWvzStellZustandBlinken(data, dataAnzeigeEigenschaftIst);

        dataAnzeigeEigenschaftIst.getScaledValue("Status").setText(statusBestimmen(anzeigePid, "tlsWzgWvzStellZustand", STATUS_OK));
      }
      else // keine Daten vorhanden
      {
        tlsWzgWvzStellZustandBlinken(null, dataAnzeigeEigenschaftIst);

        dataAnzeigeEigenschaftIst.getScaledValue("Status").setText(statusBestimmen(anzeigePid, "tlsWzgWvzStellZustand", STATUS_UNDEFINIERT));
      }
    }

    return dataAnzeigeEigenschaftIst;
  }

  /**
   * Abarbeitung des TLS-WvzStellZustand nach TLS 2002.
   *
   * @param result Attributgruppe atg.tlsWzgStellZustand.
   * @return Data
   */
  public Data tlsWzgStellZustand(ResultData result)
  {
//    System.out.println("AnzeigeTypUmsetzungTlsDefault.tlsWzgStellZustand()");

    SystemObject object = result.getObject();

    AnzeigeAktorTls anzeigeAktor = (AnzeigeAktorTls)getListeAnzeigeAktorTlsDes().get(object.getPid());

    if (anzeigeAktor == null)
      return null;

    String anzeigePid = anzeigeAktor.getAnzeige().getPid();

    Data dataAnzeigeEigenschaftIst = (Data)getListeAnzeigeEigenschaftIst().get(anzeigePid);

    m_debug.finer("tlsWzgStellZustand DE " + object.getPid() + "\nAnzeige " + anzeigePid + "\nresult " + result + "\nData " + dataAnzeigeEigenschaftIst);

    if (result.isNoSourceAvailable()) // keine Quelle vorhanden
    {
      ZwischenschichtAq.atgAnzeigeEigenschaftIstInitialisieren(dataAnzeigeEigenschaftIst, result.getDataTime(), statusBestimmen(anzeigePid, "tlsWzgStellZustand", STATUS_KOMMUNIKATIONSAUSFALL));
    }
    else // Quelle vorhanden
    {
      if (result.hasData()) // Daten vorhanden
      {
        Data data = result.getData();

        AnzeigeTypUmsetzungTls anzeigeTypUmsetzungTls = anzeigeAktor.getAnzeigeTypUmsetzung();

        String prinzip = data.getScaledValue("AnzeigePrinzip").getValueText();

        if ((prinzip.equals("a")) || (prinzip.equals("b"))) // WVZAnzeige
        {
          HashMap mapWvzInhalt = null;

          mapWvzInhalt = anzeigeTypUmsetzungTls.getListeWvzInhaltUmsetzungTlsCode();

          if (data.getArray("WVZAnzeige").getLength() == 1)
          {
            int code = data.getArray("WVZAnzeige").getItem(0).getScaledValue("Code").intValue();

            m_debug.finest("Code " + code);

            WvzInhaltUmsetzungTls wvzInhaltUmsetzungTls = (WvzInhaltUmsetzungTls)mapWvzInhalt.get(code);

            if (wvzInhaltUmsetzungTls != null)
            {
              dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getReferenceValue("AnzeigeInhalt").setText(wvzInhaltUmsetzungTls.getPidWvzInhalt());
              dataAnzeigeEigenschaftIst.getScaledValue("Status").setText(statusBestimmen(anzeigePid, "tlsWzgStellZustand", STATUS_OK));
            }
            else
            {
              dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getReferenceValue("AnzeigeInhalt").setSystemObject(null);
              dataAnzeigeEigenschaftIst.getScaledValue("Status").setText(statusBestimmen(anzeigePid, "tlsWzgStellZustand", STATUS_UNDEFINIERT));
            }

            tlsWzgWvzStellZustandBlinken(data.getArray("WVZAnzeige").getItem(0), dataAnzeigeEigenschaftIst);
          }
          else
          {
            tlsWzgWvzStellZustandBlinken(null, dataAnzeigeEigenschaftIst);

            dataAnzeigeEigenschaftIst.getScaledValue("Status").setText(statusBestimmen(anzeigePid, "tlsWzgStellZustand", STATUS_UNDEFINIERT));
          }
        }
        else
        {
          if (data.getArray("StellCode").getLength() == 1)
          {
            int code = data.getArray("StellCode").getItem(0).getScaledValue("Code").intValue();

            m_debug.finest("Code " + code);

            if ((prinzip.equals("c")) || (prinzip.equals("d")))
            {
              if (code == 0) // freie TextDefinition
              {

                // Ergebnis löschen

                dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getItem("DynamischerWechseltext").getReferenceValue("Font").setSystemObject(null);
                dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getItem("DynamischerWechseltext").getUnscaledValue("TextAusrichtung").set(-1);
                dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getItem("DynamischerWechseltext").getReferenceArray("Textdefinition").setLength(0);

                // Grundfont bestimmen

                String pidFontUmsetzungTls = anzeigeTypUmsetzungTls.getPidGrundFont();

//                if (pidFontUmsetzungTls == null)
//                {
//                  pidFontUmsetzungTls = "fontUmsetzungTls.wzg.led";
//                }

                String zeichenfolge = "";

                for (int i = 0; i < data.getArray("TextDefinition").getLength(); i++)
                {
                  zeichenfolge += data.getArray("TextDefinition").getItem(i).getTextValue("Text").getValueText(); //.getText().substring(0, 1));
                }

                int laengeInput = zeichenfolge.length();
                int laengeText = laengeInput;

                if (laengeInput == 0) // Anzeige ausgeschaltet
                {
                  // nichts weiter zu tun
                }
                else
                {
                  if (laengeInput >= 1)
                  {
                    FontUmsetzungTls fontUmsetzungTls = null;

                    if (pidFontUmsetzungTls != null)
                    {
                      HashMap mapFont = anzeigeTypUmsetzungTls.getListeFontUmsetzungTls();

                      if (mapFont != null)
                        fontUmsetzungTls = (FontUmsetzungTls)mapFont.get(pidFontUmsetzungTls);
                    }

                    HashMap mapZeichen = null;

                    if (fontUmsetzungTls != null)
                      mapZeichen = fontUmsetzungTls.getListeZeichenUmsetzungTlsCode();

                    if ((mapZeichen == null) || (mapZeichen.size() == 0))
                      mapZeichen = anzeigeTypUmsetzungTls.getListeZeichenUmsetzungTlsCode();

  //                Pid Font über Umsetzung bestimmen

                    String pidFont = null;

                    if (fontUmsetzungTls != null)
                    {
                      pidFont = fontUmsetzungTls.getPidFont();
                    }

                    if (pidFont != null)
                    {
                      dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getItem("DynamischerWechseltext").getReferenceValue("Font").setText(pidFont);
                    }

                    // Textausrichtung bestimmen

                    String textAusrichtung = anzeigeTypUmsetzungTls.getGrundTextAusrichtung();

                    dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getItem("DynamischerWechseltext").getUnscaledValue("TextAusrichtung").setText(textAusrichtung);

                    // Text

                    dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getItem("DynamischerWechseltext").getReferenceArray("Textdefinition").setLength(laengeText);

                    for (int i = 0; i < zeichenfolge.length(); i++)
                    {
                      m_debug.finest("Zeichen " + i + " " + zeichenfolge.codePointAt(i));

                      int codeZeichen = zeichenfolge.codePointAt(i);

                      ZeichenUmsetzungTls zeichenUmsetzungTls = (ZeichenUmsetzungTls)mapZeichen.get(codeZeichen);

                      if (zeichenUmsetzungTls != null)
                        dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getItem("DynamischerWechseltext").getReferenceArray("Textdefinition").getReferenceValue(i).setText(zeichenUmsetzungTls.getPidZeichen());
                      else
                        dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getItem("DynamischerWechseltext").getReferenceArray("Textdefinition").getReferenceValue(i).setSystemObject(null);
                    }
                  }

                  tlsWzgWechselTextBlinken(data.getArray("StellCode").getItem(0), dataAnzeigeEigenschaftIst);

                  dataAnzeigeEigenschaftIst.getScaledValue("Status").setText(statusBestimmen(anzeigePid, "tlsWzgStellZustand", STATUS_OK));
                }
              }
              else // code != 0
              {
                HashMap mapWvzInhalt = null;

                mapWvzInhalt = anzeigeTypUmsetzungTls.getListeWvzInhaltUmsetzungTlsCode();

                WvzInhaltUmsetzungTls wvzInhaltUmsetzungTls = (WvzInhaltUmsetzungTls)mapWvzInhalt.get(code);

                if (wvzInhaltUmsetzungTls != null)
                {
                  dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getReferenceValue("AnzeigeInhalt").setText(wvzInhaltUmsetzungTls.getPidWvzInhalt());
                }

                tlsWzgWechselTextBlinken(data.getArray("StellCode").getItem(0), dataAnzeigeEigenschaftIst);

                dataAnzeigeEigenschaftIst.getScaledValue("Status").setText(statusBestimmen(anzeigePid, "tlsWzgStellZustand", STATUS_OK));

                // Wechseltextinformationen löschen

                dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getReferenceValue("Wechseltext").setSystemObject(null);
                dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getItem("DynamischerWechseltext").getReferenceValue("Font").setSystemObject(null);
                dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getItem("DynamischerWechseltext").getUnscaledValue("TextAusrichtung").set(-1);
                dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getItem("DynamischerWechseltext").getReferenceArray("Textdefinition").setLength(0);
              }
            }
          }
          else
          {
            tlsWzgWechselTextBlinken(null, dataAnzeigeEigenschaftIst);

            dataAnzeigeEigenschaftIst.getScaledValue("Status").setText(statusBestimmen(anzeigePid, "tlsWzgStellZustand", STATUS_UNDEFINIERT));
          }
        }
      }
      else // keine Daten vorhanden
      {
        tlsWzgWechselTextBlinken(null, dataAnzeigeEigenschaftIst);

        dataAnzeigeEigenschaftIst.getScaledValue("Status").setText(statusBestimmen(anzeigePid, "tlsWzgStellZustand", STATUS_UNDEFINIERT));
      }
    }

    return dataAnzeigeEigenschaftIst;
  }

  /**
   * Abarbeitung der Befehle aus dem Modell AnzeigenGlobal.
   *
   * @param resultData AttributGruppe atg.anzeigeEigenschaft.
   * @param zeitpunktUebergeordnet Zeitpunkt der übergeordneten Schaltung.
   * @return Befehl für das Modell KExTls.
   */
  public ResultData anzeigeEigenschaftSoll(ResultData resultData, Long zeitpunktUebergeordnet)
  {
    ResultData ergebnis = null;

    SystemObject objekt = resultData.getObject();

    AnzeigeAktorTls anzeigeAktor = (AnzeigeAktorTls)getListeAnzeigeAktorTlsAnzeige().get(objekt.getPid());

    if (anzeigeAktor == null)
      return null;

    if (resultData.isNoSourceAvailable()) // keine Quelle vorhanden
    {
      return null;
    }
    else // Quelle vorhanden
    {
      if (resultData.hasData()) // Daten vorhanden
      {
        Data data = resultData.getData();

        AnzeigeTypUmsetzungTls anzeigeTypUmsetzungTls = anzeigeAktor.getAnzeigeTypUmsetzung();

        long zeitpunkt = data.getTimeValue("SollZeitpunkt").getMillis();

        // BefehlsArt prüfen

        int befehlsArt = AnzeigeEigenschaftSollBefehlsArt.befehlsArtBestimmen(data);

        if (befehlsArt == AnzeigeEigenschaftSollBefehlsArt.WvzInhalt)
        {
          if ((!isKannFreieGrafik()) && (!isKannFreiText())) // reine Stellcode-Ansteuerung ok
          {

          }
          else
          {
            // TODO WvzInhalt über Wechseltext !

            m_debug.error("Anzeige " + resultData.getObject().getPid() + " kann keine Stellcode-Ansteuerung");

            return null;
          }
        }
        else if (befehlsArt == AnzeigeEigenschaftSollBefehlsArt.Grafik)
        {
          if (isKannFreieGrafik()) // Grafik-Ansteuerung ok
          {

          }
          else
          {
            m_debug.error("Anzeige " + resultData.getObject().getPid() + " kann keine Grafik-Ansteuerung");

            return null;
          }
        }
        else if (befehlsArt == AnzeigeEigenschaftSollBefehlsArt.Wechseltext)
        {
          if (isKannFreiText()) // Wechseltext-Ansteuerung ok
          {

          }
          else
          {
            m_debug.error("Anzeige " + resultData.getObject().getPid() + " kann keine Wechseltext-Ansteuerung");

            return null;
          }
        }
        else
        {
          m_debug.error("ATG für Anzeige " + resultData.getObject().getPid() + " beinhaltet keinen Befehl");

          return null;
        }

        if (befehlsArt == AnzeigeEigenschaftSollBefehlsArt.WvzInhalt) // reine Stellcode-Ansteuerung
        {
          HashMap mapWvzInhalt = null;

          ConfigurationObject objektAnzeigeInhalt = (ConfigurationObject)data.getItem("Eigenschaft").getReferenceValue("AnzeigeInhalt").getSystemObject();

          AttributeGroup atgWvzInhalt = ZwischenschichtAq.getConnection().getDataModel().getAttributeGroup("atg.wvzInhalt");

          Data dataWvzInhalt = objektAnzeigeInhalt.getConfigurationData(atgWvzInhalt);
          String anzeigeInhalt = dataWvzInhalt.getTextValue("Bildinhalt").getValueText();

          m_debug.finest("AnzeigeInhalt " + anzeigeInhalt);

          // DEs bestimmen

          int anzahlDes = anzeigeAktor.getDes().size();

          if (anzahlDes == 1)
          {
            mapWvzInhalt = anzeigeTypUmsetzungTls.getListeWvzInhaltUmsetzungTlsWvzInhalt();

            WvzInhaltUmsetzungTls wvzInhaltUmsetzungTls = (WvzInhaltUmsetzungTls)mapWvzInhalt.get(anzeigeInhalt);

            int code = -1;

            if (wvzInhaltUmsetzungTls != null)
            {
              code = wvzInhaltUmsetzungTls.getCode();
            }
            else
            {
              m_debug.error("Objekt " + resultData.getObject().getPid() + " WvzInhalt " + anzeigeInhalt + " nicht gefunden");
            }

            Set set = anzeigeAktor.getDes().keySet();

            Iterator it = set.iterator();

            String key = (String)it.next();

            SystemObject de = anzeigeAktor.getDes().get(key);

            int anzahlGlobalUmsetzung = anzeigeTypUmsetzungTls.getListeGlobalUmsetzung().size();

            for (int j = 0; j < anzahlGlobalUmsetzung; j++)
            {
              GlobalUmsetzung globalUmsetzung = (GlobalUmsetzung)anzeigeTypUmsetzungTls.getListeGlobalUmsetzung().get(j);

              for (int k = 0; k < globalUmsetzung.getVerwendeteAttributGruppenAspekte().getAttributGruppenAspekte().size(); k++)
              {
                GlobalUmsetzungVerwendeteAttributGruppenAspekte guvaga = globalUmsetzung.getVerwendeteAttributGruppenAspekte();

                String atg = guvaga.getAttributGruppenAspekte().get(k).getPidAttributGruppe();
                String asp = guvaga.getAttributGruppenAspekte().get(k).getPidSendAspekt();
                String anzeigePrinzip = guvaga.getAttributGruppenAspekte().get(k).getAnzeigePrinzip();

                if (asp.length() > 0)
                {
                  if ((atg.equals("atg.tlsWzgWvzStellZustand")) ||
                      (atg.equals("atg.tlsWzgStellZustand")))
                  {
                    AttributeGroup atgErgebnis = ZwischenschichtAq.getConnection().getDataModel().getAttributeGroup(atg);
                    Aspect aspErgebnis = ZwischenschichtAq.getConnection().getDataModel().getAspect(asp);

                    Data dataErgebnis = ZwischenschichtAq.getConnection().createData(ZwischenschichtAq.getConnection().getDataModel().getAttributeGroup(atg));

                    if (atg.equals("atg.tlsWzgWvzStellZustand"))
                    {
                      dataErgebnis.getUnscaledValue("WvzCode").set(code);

                      // BlinkenStatus

                      String blinkenStatus = data.getItem("Eigenschaft").getScaledValue("Blinken").getValueText().equals("Ja") ? "ein" : "aus";
                      dataErgebnis.getItem("Blinken").getScaledValue("Status").setText(blinkenStatus);

                      // BlinkenDauer

                      if (blinkenStatus.equals("ein"))
                      {
                        if (data.getItem("Eigenschaft").getScaledValue("Blinktakt").isState())
                        {
                          dataErgebnis.getItem("Blinken").getScaledValue("Dauer").set(5); // Default
                        }
                        else
                        {
                          double blinkenDauerAnzeige = data.getItem("Eigenschaft").getScaledValue("Blinktakt").doubleValue();
                          if (blinkenDauerAnzeige < 1.0)
                          {
                            blinkenDauerAnzeige = 1.0;
                          }
                          if (blinkenDauerAnzeige > 3.0)
                          {
                            blinkenDauerAnzeige = 3.0;
                          }
                          int blinkenDauerTls = (int)(blinkenDauerAnzeige*10)/2;

                          dataErgebnis.getItem("Blinken").getScaledValue("Dauer").set(blinkenDauerTls);
                        }
                      }
                      else
                      {
                        dataErgebnis.getItem("Blinken").getScaledValue("Dauer").setText("aus");
                      }

                      if (zeitpunkt < ((new Date()).getTime()-600000000))
                      {
                        zeitpunkt = zeitpunktUebergeordnet.longValue();
                      }

                      ergebnis = new ResultData(de, new DataDescription(atgErgebnis, aspErgebnis), zeitpunkt, dataErgebnis);
                    }

                    if (atg.equals("atg.tlsWzgStellZustand"))
                    {
                      dataErgebnis.getScaledValue("AnzeigePrinzip").setText(anzeigePrinzip);

                      if ((anzeigePrinzip.equals("a")) || (anzeigePrinzip.equals("b")))
                      {
                        dataErgebnis.getArray("WVZAnzeige").setLength(1);
                        dataErgebnis.getArray("StellCode").setLength(0);
                      }
                      else
                      {
                        dataErgebnis.getArray("WVZAnzeige").setLength(0);
                        dataErgebnis.getArray("StellCode").setLength(1);
                      }

                      dataErgebnis.getArray("TextDefinition").setLength(0);
                      dataErgebnis.getArray("GruppenDefinition").setLength(0);

                      if ((anzeigePrinzip.equals("a")) || (anzeigePrinzip.equals("b")))
                      {
                        dataErgebnis.getArray("WVZAnzeige").getItem(0).getUnscaledValue("Code").set(code);

                        // BlinkenStatus

                        String blinkenStatus = data.getItem("Eigenschaft").getScaledValue("Blinken").getValueText().equals("Ja") ? "ein" : "aus";
                        dataErgebnis.getArray("WVZAnzeige").getItem(0).getItem("Blinken").getScaledValue("Status").setText(blinkenStatus);

                        // BlinkenDauer

                        if (blinkenStatus.equals("ein"))
                        {
                          if (data.getItem("Eigenschaft").getScaledValue("Blinktakt").isState())
                          {
                            dataErgebnis.getArray("WVZAnzeige").getItem(0).getItem("Blinken").getScaledValue("Dauer").set(5); // Default
                          }
                          else
                          {
                            double blinkenDauerAnzeige = data.getItem("Eigenschaft").getScaledValue("Blinktakt").doubleValue();
                            if (blinkenDauerAnzeige < 1.0)
                            {
                              blinkenDauerAnzeige = 1.0;
                            }
                            if (blinkenDauerAnzeige > 3.0)
                            {
                              blinkenDauerAnzeige = 3.0;
                            }
                            int blinkenDauerTls = (int)(blinkenDauerAnzeige*10)/2;

                            dataErgebnis.getArray("WVZAnzeige").getItem(0).getItem("Blinken").getScaledValue("Dauer").set(blinkenDauerTls);
                          }
                        }
                        else
                        {
                          dataErgebnis.getArray("WVZAnzeige").getItem(0).getItem("Blinken").getScaledValue("Dauer").setText("aus");
                        }

                        dataErgebnis.getArray("WVZAnzeige").getItem(0).getItem("Status").getScaledValue("Fehler").setText("ok");
                        dataErgebnis.getArray("WVZAnzeige").getItem(0).getItem("Status").getScaledValue("Programm").setText("abgeschlossen");
                      }
                      else
                      {
                        dataErgebnis.getArray("StellCode").getItem(0).getUnscaledValue("Code").set(code);

                        // BlinkenStatus

                        String blinkenStatus = data.getItem("Eigenschaft").getScaledValue("Blinken").getValueText().equals("Ja") ? "ein" : "aus";
                        dataErgebnis.getArray("StellCode").getItem(0).getItem("Blinken").getScaledValue("Status").setText(blinkenStatus);

                        // BlinkenDauer

                        if (blinkenStatus.equals("ein"))
                        {
                          if (data.getItem("Eigenschaft").getScaledValue("Blinktakt").isState())
                          {
                            dataErgebnis.getArray("StellCode").getItem(0).getItem("Blinken").getScaledValue("Dauer").set(5); // Default
                          }
                          else
                          {
                            double blinkenDauerAnzeige = data.getItem("Eigenschaft").getScaledValue("Blinktakt").doubleValue();
                            if (blinkenDauerAnzeige < 1.0)
                            {
                              blinkenDauerAnzeige = 1.0;
                            }
                            if (blinkenDauerAnzeige > 3.0)
                            {
                              blinkenDauerAnzeige = 3.0;
                            }
                            int blinkenDauerTls = (int)(blinkenDauerAnzeige*10)/2;

                            dataErgebnis.getArray("StellCode").getItem(0).getItem("Blinken").getScaledValue("Dauer").set(blinkenDauerTls);
                          }
                        }
                        else
                        {
                          dataErgebnis.getArray("StellCode").getItem(0).getItem("Blinken").getScaledValue("Dauer").setText("aus");
                        }

                        dataErgebnis.getArray("StellCode").getItem(0).getItem("Status").getScaledValue("Fehler").setText("ok");
                        dataErgebnis.getArray("StellCode").getItem(0).getItem("Status").getScaledValue("Programm").setText("abgeschlossen");
                      }

                      if (zeitpunkt < ((new Date()).getTime()-600000000))
                      {
                        zeitpunkt = zeitpunktUebergeordnet.longValue();
                      }

                      ergebnis = new ResultData(de, new DataDescription(atgErgebnis, aspErgebnis), zeitpunkt, dataErgebnis);
                    }

                    break;
                  }
                }
              }
            }
          }
        }
        else
        {
          if (befehlsArt == AnzeigeEigenschaftSollBefehlsArt.Wechseltext) // reine Textansteuerung
          {
            ConfigurationObject wechseltext = (ConfigurationObject)data.getItem("Eigenschaft").getReferenceValue("Wechseltext").getSystemObject();

            Data dataWechseltext = null;

            if (wechseltext != null) // konfigurierter Wechseltext
            {
              AttributeGroup atgWechseltext = ZwischenschichtAq.getConnection().getDataModel().getAttributeGroup("atg.wechseltext");
              dataWechseltext = wechseltext.getConfigurationData(atgWechseltext);
            }
            else // dynamischer Wechseltext
            {
              dataWechseltext = data.getItem("Eigenschaft").getItem("DynamischerWechseltext");
            }

            String pidFont = dataWechseltext.getReferenceValue("Font").getValueText();

            HashMap mapZeichen = null;

            HashMap<String, FontUmsetzungTls> mapFont = anzeigeTypUmsetzungTls.getListeFontUmsetzungTls();

            if ((pidFont != null) && (mapFont != null))
            {
              TreeSet<String> treeset = new TreeSet<String>();
              treeset.addAll(mapFont.keySet());

              Iterator iterator = treeset.iterator();

              FontUmsetzungTls fontUmsetzungTls;

              while (iterator.hasNext())
              {
                fontUmsetzungTls = (FontUmsetzungTls)mapFont.get(iterator.next());

                if (fontUmsetzungTls.getPidFont().equals(pidFont))
                {
                  mapZeichen = fontUmsetzungTls.getListeZeichenUmsetzungTlsZeichen();

                  break;
                }
              }
            }

            // DEs bestimmen

            int anzahlDes = anzeigeAktor.getDes().size();

            if (anzahlDes == 1) // Wechseltext nur mit einer DE
            {
              Set set = anzeigeAktor.getDes().keySet();

              Iterator it = set.iterator();

              String key = (String)it.next();

              SystemObject de = anzeigeAktor.getDes().get(key);

              int anzahlGlobalUmsetzung = anzeigeTypUmsetzungTls.getListeGlobalUmsetzung().size();

              for (int j = 0; j < anzahlGlobalUmsetzung; j++)
              {
                GlobalUmsetzung globalUmsetzung = (GlobalUmsetzung)anzeigeTypUmsetzungTls.getListeGlobalUmsetzung().get(j);

                for (int k = 0; k < globalUmsetzung.getVerwendeteAttributGruppenAspekte().getAttributGruppenAspekte().size(); k++)
                {
                  GlobalUmsetzungVerwendeteAttributGruppenAspekte guvaga = globalUmsetzung.getVerwendeteAttributGruppenAspekte();

                  String atg = guvaga.getAttributGruppenAspekte().get(k).getPidAttributGruppe();
                  String asp = guvaga.getAttributGruppenAspekte().get(k).getPidSendAspekt();
                  String anzeigePrinzip = guvaga.getAttributGruppenAspekte().get(k).getAnzeigePrinzip();

                  if (asp.length() > 0)
                  {
                    if ((atg.equals("atg.tlsWzgWechselText")) ||
                        (atg.equals("atg.tlsWzgStellZustand")))
                    {
                      AttributeGroup atgErgebnis = ZwischenschichtAq.getConnection().getDataModel().getAttributeGroup(atg);
                      Aspect aspErgebnis = ZwischenschichtAq.getConnection().getDataModel().getAspect(asp);

                      Data dataErgebnis = ZwischenschichtAq.getConnection().createData(ZwischenschichtAq.getConnection().getDataModel().getAttributeGroup(atg));

                      if (atg.equals("atg.tlsWzgWechselText"))
                      {
                        StringBuffer befehl = new StringBuffer();

                        ReferenceValue[] text = dataWechseltext.getReferenceArray("Textdefinition").getReferenceValues();

                        for (int i = 0; i < text.length; i++)
                        {
                          String pid = text[i].getSystemObject().getPid();

                          if (mapZeichen == null)
                          {
                            continue;
                          }

                          ZeichenUmsetzungTls zeichenUmsetzungTls = (ZeichenUmsetzungTls)mapZeichen.get(pid);

                          if (zeichenUmsetzungTls != null)
                          {
                            befehl.append(String.format("%c", zeichenUmsetzungTls.getZeichencode()));
                          }
                          else
                          {
                            befehl.append(String.format("%c", 32)); // Leerzeichen als default
                          }
                        }

                        dataErgebnis.getTextValue("WVZTextZeichen").setText(befehl.toString());

                        // BlinkenStatus

                        String blinkenStatus = data.getItem("Eigenschaft").getScaledValue("Blinken").getValueText().equals("Ja") ? "ein" : "aus";
                        dataErgebnis.getItem("Blinken").getScaledValue("Status").setText(blinkenStatus);

                        // BlinkenDauer

                        if (blinkenStatus.equals("ein"))
                        {
                          if (data.getItem("Eigenschaft").getScaledValue("Blinktakt").isState())
                          {
                            dataErgebnis.getItem("Blinken").getScaledValue("Dauer").set(5); // Default
                          }
                          else
                          {
                            double blinkenDauerAnzeige = data.getItem("Eigenschaft").getScaledValue("Blinktakt").doubleValue();
                            if (blinkenDauerAnzeige < 1.0)
                            {
                              blinkenDauerAnzeige = 1.0;
                            }
                            if (blinkenDauerAnzeige > 3.0)
                            {
                              blinkenDauerAnzeige = 3.0;
                            }
                            int blinkenDauerTls = (int)(blinkenDauerAnzeige*10)/2;

                            dataErgebnis.getItem("Blinken").getScaledValue("Dauer").set(blinkenDauerTls);
                          }
                        }
                        else
                        {
                          dataErgebnis.getItem("Blinken").getScaledValue("Dauer").setText("aus");
                        }

                        if (zeitpunkt < ((new Date()).getTime()-600000000))
                        {
                          zeitpunkt = zeitpunktUebergeordnet.longValue();
                        }

                        ergebnis = new ResultData(de, new DataDescription(atgErgebnis, aspErgebnis), zeitpunkt, dataErgebnis);
                      }

                      if (atg.equals("atg.tlsWzgStellZustand"))
                      {
                        dataErgebnis.getScaledValue("AnzeigePrinzip").setText(anzeigePrinzip);
                        dataErgebnis.getArray("WVZAnzeige").setLength(0);
                        dataErgebnis.getArray("StellCode").setLength(1);
                        dataErgebnis.getArray("TextDefinition").setLength(0);
                        dataErgebnis.getArray("GruppenDefinition").setLength(0);

                        dataErgebnis.getArray("StellCode").getItem(0).getUnscaledValue("Code").set(0);

                        // BlinkenStatus

                        String blinkenStatus = data.getItem("Eigenschaft").getScaledValue("Blinken").getValueText().equals("Ja") ? "ein" : "aus";
                        dataErgebnis.getArray("StellCode").getItem(0).getItem("Blinken").getScaledValue("Status").setText(blinkenStatus);

                        // BlinkenDauer

                        if (blinkenStatus.equals("ein"))
                        {
                          if (data.getItem("Eigenschaft").getScaledValue("Blinktakt").isState())
                          {
                            dataErgebnis.getArray("StellCode").getItem(0).getItem("Blinken").getScaledValue("Dauer").set(5); // Default
                          }
                          else
                          {
                            double blinkenDauerAnzeige = data.getItem("Eigenschaft").getScaledValue("Blinktakt").doubleValue();
                            if (blinkenDauerAnzeige < 1.0)
                            {
                              blinkenDauerAnzeige = 1.0;
                            }
                            if (blinkenDauerAnzeige > 3.0)
                            {
                              blinkenDauerAnzeige = 3.0;
                            }
                            int blinkenDauerTls = (int)(blinkenDauerAnzeige*10)/2;

                            dataErgebnis.getArray("StellCode").getItem(0).getItem("Blinken").getScaledValue("Dauer").set(blinkenDauerTls);
                          }
                        }
                        else
                        {
                          dataErgebnis.getArray("StellCode").getItem(0).getItem("Blinken").getScaledValue("Dauer").setText("aus");
                        }

                        dataErgebnis.getArray("StellCode").getItem(0).getItem("Status").getScaledValue("Fehler").setText("ok");
                        dataErgebnis.getArray("StellCode").getItem(0).getItem("Status").getScaledValue("Programm").setText("abgeschlossen");

                        ReferenceValue[] text = dataWechseltext.getReferenceArray("Textdefinition").getReferenceValues();

                        dataErgebnis.getArray("TextDefinition").setLength(text.length);

                        int index = 0;

                        int zeichen;

                        for (int i = 0; i < text.length; i++)
                        {
                          String pid = text[i].getSystemObject().getPid();

                          zeichen = 32; // Leerzeichen als default

                          if (mapZeichen != null)
                          {
                            ZeichenUmsetzungTls zeichenUmsetzungTls = (ZeichenUmsetzungTls)mapZeichen.get(pid);

                            if (zeichenUmsetzungTls != null)
                            {
                              zeichen = zeichenUmsetzungTls.getZeichencode();
                            }
                            else
                            {
                              zeichen = 32; // Leerzeichen als default
                            }
                          }

                          dataErgebnis.getArray("TextDefinition").getItem(index++).getTextValue("Text").setText(String.format("%c", zeichen).toString());
                        }

                        if (zeitpunkt < ((new Date()).getTime()-600000000))
                        {
                          zeitpunkt = zeitpunktUebergeordnet.longValue();
                        }

                        ergebnis = new ResultData(de, new DataDescription(atgErgebnis, aspErgebnis), zeitpunkt, dataErgebnis);
                      }

                      break;
                    }
                  }
                }
              }
            }
            else // anzahlDes != 1
            {
              m_debug.error("Wechseltext nur mit einer DE möglich");
            }
          }
        }
      }
      else // keine Daten vorhanden
      {
        return null;
      }
    }

    return ergebnis;
  }

  /**
   * Abarbeitung des Anteiles Blinken des TLS-WechselText genau nach TLS 93.
   *
   * @param data Daten tlsWzgBlinken.
   * @param dataAnzeigeEigenschaftIst Daten AnzeigeEigenschaftIst.
   * @return Daten AnzeigeEigenschaftIst mit eingearbeiteten Blinkinformation.
   */
  protected Data tlsWzgWechselTextBlinken(Data data, Data dataAnzeigeEigenschaftIst)
  {
    return tlsWzgBlinken(data, dataAnzeigeEigenschaftIst);
  }

  /**
   * Abarbeitung des Anteiles Blinken des TLS-WzgWvzStellZustand genau nach TLS 93.
   *
   * @param data Daten tlsWzgBlinken.
   * @param dataAnzeigeEigenschaftIst Daten AnzeigeEigenschaftIst.
   * @return Daten AnzeigeEigenschaftIst mit eingearbeiteten Blinkinformation.
   */
  protected Data tlsWzgWvzStellZustandBlinken(Data data, Data dataAnzeigeEigenschaftIst)
  {
    return tlsWzgBlinken(data, dataAnzeigeEigenschaftIst);
  }

  /**
   * Abarbeitung des Anteiles Blinken des TLS-WzgStellZustand genau nach TLS 93.
   *
   * @param data Daten tlsWzgBlinken.
   * @param dataAnzeigeEigenschaftIst Daten AnzeigeEigenschaftIst.
   * @return Daten AnzeigeEigenschaftIst mit eingearbeiteten Blinkinformation.
   */
  protected Data tlsWzgStellZustandBlinken(Data data, Data dataAnzeigeEigenschaftIst)
  {
    return tlsWzgBlinken(data, dataAnzeigeEigenschaftIst);
  }

  /**
   * Abarbeitung des Anteiles Blinken genau nach TLS 93.
   *
   * @param data Daten tlsWzgBlinken.
   * @param dataAnzeigeEigenschaftIst Daten AnzeigeEigenschaftIst.
   * @return Daten AnzeigeEigenschaftIst mit eingearbeiteten Blinkinformation.
   */
  protected Data tlsWzgBlinken(Data data, Data dataAnzeigeEigenschaftIst)
  {
    dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getScaledValue("Blinken").setText("Nein");
    dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getScaledValue("Blinktakt").setText("Undefiniert");

    if (data != null)
    {
      String blinkenStatusTls = data.getItem("Blinken").getTextValue("Status").getValueText();

      if (blinkenStatusTls.equals("ein"))
      {
        int blinkDauerTls = data.getItem("Blinken").getUnscaledValue("Dauer").intValue();
        int blinkDauerAq = blinkDauerTls*200; // Faktor 200 ms

        dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getScaledValue("Blinken").setText("Ja");
        dataAnzeigeEigenschaftIst.getItem("Eigenschaft").getUnscaledValue("Blinktakt").set(blinkDauerAq);
      }
    }

    return dataAnzeigeEigenschaftIst;
  }

  /**
   * Holen der HashMap mit AnzeigeAktorTls mit Key Anzeige.
   *
   * @return HashMap mit AnzeigeAktorTls mit Key Anzeige.
   */
  public HashMap getListeAnzeigeAktorTlsAnzeige()
  {
    return m_listeAnzeigeAktorTlsAnzeige;
  }

  /**
   * Holen der HashMap mit AnzeigeAktorTls mit Key DE's.
   *
   * @return HashMap mit AnzeigeAktorTls mit Key DE's.
   */
  public HashMap getListeAnzeigeAktorTlsDes()
  {
    return m_listeAnzeigeAktorTlsDes;
  }

  /**
   * Holen der HashMap mit Attributgruppe atg.anzeigeEigenschaftIst für alle Anzeigen aktuell.
   *
   * @return HashMap mit Attributgruppe atg.anzeigeEigenschaftIst für alle Anzeigen aktuell.
   */
  public HashMap getListeAnzeigeEigenschaftIst()
  {
    return m_listeAnzeigeEigenschaftIst;
  }

  /**
   * Holen der HashMap für Status Methoden für alle Anzeigen.
   *
   * @return HashMap für Status Methoden für alle Anzeigen.
   */
  public HashMap getListeAnzeigeStatus()
  {
    return m_listeAnzeigeStatus;
  }

  /**
   * Holen der Information, ob die Anzeige freie Grafik darstellen kann.
   *
   * @return <code>true</code>, wenn die Anzeige freie Grafik darstellen kann, sonst <code>false</code>.
   */
  public boolean isKannFreieGrafik()
  {
    return m_kannFreieGrafik;
  }

  /**
   * Setzen der Information, ob die Anzeige freie Grafik darstellen kann.
   *
   * @param kannFreieGrafik <code>true</code>, wenn die Anzeige freie Grafik darstellen kann, sonst <code>false</code>.
   */
  public void setKannFreieGrafik(boolean kannFreieGrafik)
  {
    m_kannFreieGrafik = kannFreieGrafik;
  }

  /**
   * Holen der Information, ob die Anzeige freien Text darstellen kann.
   *
   * @return <code>true</code>, wenn die Anzeige freien Text darstellen kann, sonst <code>false</code>.
   */
  public boolean isKannFreiText()
  {
    return m_kannFreiText;
  }

  /**
   * Setzen der Information, ob die Anzeige freien Text darstellen kann.
   *
   * @param kannFreiText <code>true</code>, wenn die Anzeige freien Text darstellen kann, sonst <code>false</code>.
   */
  public void setKannFreiText(boolean kannFreiText)
  {
    m_kannFreiText = kannFreiText;
  }

  /**
   * Bestimmen des aktuellen Anzeige-Status mit der Möglichkeit gleichzeitiger Übergabe des Status einer Abarbeitungs-Methode.
   *
   * @param anzeigePid Pid der Anzeige.
   * @param methode Abarbeitungsmethode oder <code>null</code>, wenn nur Bestimmung des aktuellen Status.
   * @param status aktuelle Status der Abarbeitungsmethode, wenn methode nicht <code>null</code>.
   * @return aktueller Anzeige-Status, bestimmt aus allen für die Anzeige vorhandenen Methoden.
   */
  @SuppressWarnings("unchecked")
  protected String statusBestimmen(String anzeigePid, String methode, int status)
  {
//    System.out.println("statusBestimmen");

    HashMap<String, Integer> statusMethoden = (HashMap<String, Integer>)getListeAnzeigeStatus().get(anzeigePid);

    if (statusMethoden == null)
    {
      statusMethoden = new HashMap<String, Integer>();

      getListeAnzeigeStatus().put(anzeigePid, statusMethoden);
    }

    if (methode != null)
    {
      boolean übernehmen = true;

      if (methode.equals("tlsWzgGestörteTextPositionen"))
      {
        if ((status != STATUS_TEILSTOERUNG) && (status != STATUS_OK))
            übernehmen = false;
      }

      if (methode.equals("tlsWzgNichtDarstellbareWVZ"))
      {
        if ((status != STATUS_TEILSTOERUNG) && (status != STATUS_OK))
            übernehmen = false;
      }

      if (übernehmen)
      {
        statusMethoden.put(methode, status);
      }
    }

    Set<String> set = statusMethoden.keySet();
    Iterator iterator = set.iterator();

    int wert = 0;
    int wertMax = STATUS_OK;

    while (iterator.hasNext())
    {
      String meth = (String)iterator.next();

      wert = statusMethoden.get(meth);

      wertMax = (wert > wertMax) ? wert : wertMax;
    }

    return statusAusWertBestimmen(wertMax);
  }

  /**
   * Umsetzung des kodierten Status in den entsprechenden Text.
   *
   * @param status kodierter Status.
   * @return dem Status entsprechender Text.
   */
  protected String statusAusWertBestimmen(int status)
  {
    String ergebnis = "OK";

    switch (status)
    {
      case STATUS_GESTOERT:
      {
        ergebnis = "Gestört";

        break;
      }

      case STATUS_KOMMUNIKATIONSAUSFALL:
      {
        ergebnis = "Kommunikationsausfall";

        break;
      }

      case STATUS_STROMAUSFALL:
      {
        ergebnis = "Stromausfall";

        break;
      }

      case STATUS_TEILSTOERUNG:
      {
        ergebnis = "Teilstörung";

        break;
      }

      case STATUS_TUERKONTAKT:
      {
        ergebnis = "Türkontakt";

        break;
      }

      case STATUS_UNDEFINIERT:
      {
        ergebnis = "Undefiniert";

        break;
      }

      default:
      {
        break;
      }
    }

    return ergebnis;
  }

  /**
   * Bestimmung der Text-Ausrichtung, kann sie nicht bestimmt werden, wird die Text-Ausrichtung 'Undefiniert' zurückgegeben.
   *
   * @param anzeigeTypUmsetzungTls Objekt vom Typ AnzeigeTypUmsetzungTls.
   * @param code Zeichen aus Text von KExTls.
   * @return Text-Ausrichtung.
   */
  private String textAusrichtungAusTlsBestimmen(AnzeigeTypUmsetzungTls anzeigeTypUmsetzungTls, int code)
  {
    String ergebnis = "Undefiniert";

    HashMap<String, TextAusrichtungUmsetzungTls> mapTextAusrichtung = anzeigeTypUmsetzungTls.getListeTextAusrichtungUmsetzungTls();

    if (mapTextAusrichtung != null)
    {
      TreeSet<String> treeset = new TreeSet<String>();
      treeset.addAll(mapTextAusrichtung.keySet());

      Iterator iterator = treeset.iterator();

      TextAusrichtungUmsetzungTls textAusrichtungUmsetzungTls;

      while (iterator.hasNext())
      {
        textAusrichtungUmsetzungTls = (TextAusrichtungUmsetzungTls)mapTextAusrichtung.get(iterator.next());

        if (code == textAusrichtungUmsetzungTls.getCodeTextAusrichtung())
        {
          ergebnis = textAusrichtungUmsetzungTls.getTextAusrichtung();

          break;
        }
      }
    }

    return ergebnis;
  }

  /**
   * Bestimmung des Code der Text-Ausrichtung, kann sie nicht bestimmt werden, wird Code 0 zurückgegeben.
   *
   * @param anzeigeTypUmsetzungTls Objekt vom Typ AnzeigeTypUmsetzungTls.
   * @param pidAusrichtung Pid der TextAusrichtung
   * @return Code der TextAusrichtung
   */
  private int textAusrichtungAusAnzeigeBestimmen(AnzeigeTypUmsetzungTls anzeigeTypUmsetzungTls, String pidAusrichtung)
  {
    int ergebnis = 133; // Linksbündig als Default

    HashMap<String, TextAusrichtungUmsetzungTls> mapTextAusrichtung = anzeigeTypUmsetzungTls.getListeTextAusrichtungUmsetzungTls();

    if ((pidAusrichtung != null) && (mapTextAusrichtung != null))
    {
      TreeSet<String> treeset = new TreeSet<String>();
      treeset.addAll(mapTextAusrichtung.keySet());

      Iterator iterator = treeset.iterator();

      TextAusrichtungUmsetzungTls textAusrichtungUmsetzungTls;

      while (iterator.hasNext())
      {
        textAusrichtungUmsetzungTls = (TextAusrichtungUmsetzungTls)mapTextAusrichtung.get(iterator.next());

        if (textAusrichtungUmsetzungTls.getTextAusrichtung().equals(pidAusrichtung))
        {
          ergebnis = textAusrichtungUmsetzungTls.getCodeTextAusrichtung();

          break;
        }
      }
    }

    return ergebnis;
  }

  /**
   * Bestimmung der Pid der FontUmsetzungTls, kann er nicht bestimmt werden, wird die Pid des GrundFontUmsetzungTls zurückgegeben.
   *
   * @param anzeigeTypUmsetzungTls Objekt vom Typ AnzeigeTypUmsetzungTls.
   * @param code Zeichen aus Text von KExTls.
   * @param pidGrundFontUmsetzungTls Pid des GrundFonts der FontUmsetzungTls.
   * @return Pid der FontUmsetzungTls.
   */
  private String fontUmsetzungAusTlsBestimmen(AnzeigeTypUmsetzungTls anzeigeTypUmsetzungTls, int code, String pidGrundFontUmsetzungTls)
  {
    String ergebnis = pidGrundFontUmsetzungTls;

    HashMap<String, FontUmsetzungTls> mapFont = anzeigeTypUmsetzungTls.getListeFontUmsetzungTls();

    if (mapFont != null)
    {
      TreeSet<String> treeset = new TreeSet<String>();
      treeset.addAll(mapFont.keySet());

      Iterator iterator = treeset.iterator();

      FontUmsetzungTls fontUmsetzungTls;

      while (iterator.hasNext())
      {
        fontUmsetzungTls = (FontUmsetzungTls)mapFont.get(iterator.next());

        if (code == fontUmsetzungTls.getCodeFont())
        {
          ergebnis = fontUmsetzungTls.getPid();

          break;
        }
      }
    }

    return ergebnis;
  }

  /**
   * Bestimmung des Code der FontUmsetzungTls, kann er nicht bestimmt werden, wird der Code 0 zurückgegeben.
   * 
   * @param anzeigeTypUmsetzungTls Objekt vom Typ AnzeigeTypUmsetzungTls.
   * @param pidFont Pid des Font
   * @return Code des Font
   */
  private int fontUmsetzungAusAnzeigeBestimmen(AnzeigeTypUmsetzungTls anzeigeTypUmsetzungTls, String pidFont)
  {
    int ergebnis = 128; // Mittelschrift als Default

    HashMap<String, FontUmsetzungTls> mapFont = anzeigeTypUmsetzungTls.getListeFontUmsetzungTls();
    
    if ((pidFont != null) && (mapFont != null))
    {
      TreeSet<String> treeset = new TreeSet<String>();
      treeset.addAll(mapFont.keySet());
    
      Iterator iterator = treeset.iterator();
    
      FontUmsetzungTls fontUmsetzungTls;
      
      while (iterator.hasNext())
      {
        fontUmsetzungTls = (FontUmsetzungTls)mapFont.get(iterator.next());
        
        if (fontUmsetzungTls.getPidFont().equals(pidFont))
        {
          ergebnis = fontUmsetzungTls.getCodeFont();
          
          break;
        }
      }
    }
    
    return ergebnis;
  }
}
