/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */

package zswzg.kex.tls.anzeigenglobal;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashMap;

import zswzg.kex.tls.AnzeigeAktorTls;
import zswzg.kex.tls.AnzeigeTypUmsetzungTls;
import zswzg.kex.tls.GlobalUmsetzung;
import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataNotSubscribedException;
import de.bsvrz.dav.daf.main.OneSubscriptionPerSendData;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Klasse für die Speicherung der Informationen eines Konfigurationsobjektes vom Typ typ.anzeige
 * und die Abarbeitung der über diesen Typ enpfangenen spezifischen Attributgruppen.
 * Die Abarbeitung erfolgt mit den versorgten Umsetzungsklassen, oder wenn keine spezielle versorgt, mit der Standardumsetzung.<br>
 * Die Standardumsetzungsklasse ist die Klasse 'zswzg.kex.tls.AnzeigeTypUmsetzungTlsDefault'.<br>
 * <br>
 * Abarbeitung: AnzeigenGlobal -&gt; KExTls.<br>
 *
 * @version $Revision: 1.3 $ / $Date: 2011/01/20 06:21:36 $ / ($Author: Drapp $)
 *
 * @author Dambach-Werke GmbH
 * @author Thomas Drapp
 *
 */
public class Anzeige implements ClientReceiverInterface, ClientSenderInterface
{
  /**
   * DebugLogger für Debug-Ausgaben.
   */
  private static Debug m_debug = Debug.getLogger();
  /**
   * Konfigurationsobjekt vom Typ typ.anzeige.
   */
  private ConfigurationObject m_anzeige;
  /**
   * Datenverteilerverbindung.
   */
  private ClientDavInterface m_connection;
  /**
   * Datenmodell.
   */
  private DataModel m_dataModel;
  /**
   * AnzeigeAktorTls mit Key DEs.
   */
  private HashMap<String, AnzeigeAktorTls> m_listeAnzeigeAktorTlsKeyDes;
  /**
   * AnzeigeAktorTls mit Key Anzeige.
   */
  private HashMap<String, AnzeigeAktorTls> m_listeAnzeigeAktorTlsKeyAnzeige;
  /**
   * Defaultklasse für die AnzeigeTypUmsetzungTls.
   */
  private String klassenNameDefault = "zswzg.kex.tls.AnzeigeTypUmsetzungTlsDefault";
  /**
   * Kombinationen von AttributGruppe und Methode für die Umsetzung der Anzeige.
   */
  private String[][] methoden =
  {
    {"atg.anzeigeEigenschaftSoll", "anzeigeEigenschaftSoll"},
  };
  /**
   * Information, ob die Anzeige freie Grafik darstellen kann.
   */
  private boolean m_kannFreieGrafik = false;
  /**
   * Information, ob die Anzeige freien Text darstellen kann.
   */
  private boolean m_kannFreiText = false;
  /**
   * Information, ob die Daten zum Versand angemeldet sind.
   */
  private boolean m_datenversandAngemeldet = false;
  /**
   * Information, ob die Daten zum Empfang angemeldet sind.
   */
  private boolean m_datenempfangAngemeldet = false;
  /**
   * Instanz der Klasse AnzeigeQuerschnitt dem die Anzeige angehört.
   */
  private AnzeigeQuerschnitt m_anzeigeQuerschnitt = null;

  /**
   * Konstruktor mit Übergabe der notwendigen Parameter.
   *
   * @param connection Datenverteilerverbindung.
   * @param anzeige Konfigurationsobjekt vom Typ typ.anzeige.
   * @param listeAnzeigeAktorTlsKeyDes HashMap mit den AnzeigeAktorTls mit Key DEs.
   * @param listeAnzeigeAktorTlsKeyAnzeige HashMap mit den AnzeigeAktorTls mit Key Anzeige.
   */
  public Anzeige(ClientDavInterface connection, ConfigurationObject anzeige, HashMap<String, AnzeigeAktorTls> listeAnzeigeAktorTlsKeyDes, HashMap<String, AnzeigeAktorTls> listeAnzeigeAktorTlsKeyAnzeige)
  {
    AttributeGroup atgDaten;

    m_connection = connection;
    m_anzeige = anzeige;
    m_listeAnzeigeAktorTlsKeyDes = listeAnzeigeAktorTlsKeyDes;
    m_listeAnzeigeAktorTlsKeyAnzeige = listeAnzeigeAktorTlsKeyAnzeige;
    m_dataModel = m_connection.getDataModel();

//    m_debug.error("Anzeige " + m_anzeige.getPid());

    // Konfigurationsdaten holen

    atgDaten = m_dataModel.getAttributeGroup("atg.anzeige");

    Data daten = anzeige.getConfigurationData(atgDaten);

    SystemObject anzeigeTyp = daten.getReferenceValue("AnzeigeTyp").getSystemObject();

    if (anzeigeTyp != null)
    {
      atgDaten = m_dataModel.getAttributeGroup("atg.anzeigeTyp");

      daten = anzeigeTyp.getConfigurationData(atgDaten);

      setKannFreieGrafik(daten.getScaledValue("KannFreieGrafik").getValueText().equals("Ja"));
      setKannFreiText(daten.getScaledValue("KannFreiText").getValueText().equals("Ja"));
    }

//    anmeldungDatenempfang(this);
//
//    anmeldungDatenversand(this);
  }

  /**
   * Anmelden für den Datenversand.<br>
   * Die Anmeldung wird nur durchgeführt, wenn nicht angemeldet.<br>
   *
   * @param sender Instanz des Senders.
   */
  protected void anmeldungDatenversand(ClientSenderInterface sender)
  {
    if (m_datenversandAngemeldet)
    {
      return;
    }

    // Anzeige auf Datenversand anmelden

    AttributeGroup atgDaten = m_dataModel.getAttributeGroup("atg.anzeigeEigenschaftIst");
    Aspect aspDaten = m_dataModel.getAspect("asp.aqZustand");
    DataDescription ddDaten = new DataDescription(atgDaten, aspDaten);

    m_debug.config("Anmeldung Quelle " + m_anzeige + " " + ddDaten);

    try
    {
      m_connection.subscribeSender(sender, m_anzeige, ddDaten, SenderRole.source());
    }
    catch (OneSubscriptionPerSendData e)
    {
      e.printStackTrace();

      System.exit(0);
    }

    m_datenversandAngemeldet = true;
  }

  /**
   * Abmelden vom Datenversand.<br>
   * Die Abmeldung wird nur durchgeführt, wenn angemeldet.<br>
   *
   * @param sender Instanz des Senders.
   */
  protected void abmeldungDatenversand(ClientSenderInterface sender)
  {
    if (!m_datenversandAngemeldet)
    {
      return;
    }

    AttributeGroup atgDaten;
    Aspect aspDaten;
    DataDescription ddDaten;

    // Anzeige vom Datenversand anmelden

    atgDaten = m_dataModel.getAttributeGroup("atg.anzeigeEigenschaftIst");
    aspDaten = m_dataModel.getAspect("asp.aqZustand");
    ddDaten = new DataDescription(atgDaten, aspDaten);

    m_debug.config("Abmeldung " + m_anzeige + " " + ddDaten);

    m_connection.unsubscribeSender(sender, m_anzeige, ddDaten);

    m_datenversandAngemeldet = false;
  }

  /**
   * Anmelden für den Datenempfang.<br>
   * Die Anmeldung wird nur durchgeführt, wenn nicht angemeldet.<br>
   *
   * @param empfaenger Instanz des Empfängers.
   */
  protected void anmeldungDatenempfang(ClientReceiverInterface empfaenger)
  {
    if (m_datenempfangAngemeldet)
    {
      return;
    }

    AttributeGroup atgDaten;
    Aspect aspDaten;
    DataDescription ddDaten;

    // Anzeige auf Datenempfang anmelden

    atgDaten = m_dataModel.getAttributeGroup("atg.anzeigeEigenschaftSoll");
    aspDaten = m_dataModel.getAspect("asp.aqVorgabe");
    ddDaten = new DataDescription(atgDaten, aspDaten);

    m_debug.config("Anmeldung Senke " + m_anzeige.getPid() + " " + ddDaten);

    try
    {
      m_connection.subscribeReceiver(empfaenger, m_anzeige, ddDaten, ReceiveOptions.normal(), ReceiverRole.drain());
    }
    catch (NullPointerException e)
    {
      e.printStackTrace();

      System.exit(0);
    }
  }

  /**
   * Abmelden vom Datenempfang.<br>
   * Die Abmeldung wird nur durchgeführt, wenn angemeldet.<br>
   *
   * @param empfaenger Instanz des Empfängers.
   */
  protected void abmeldungDatenempfang(ClientReceiverInterface empfaenger)
  {
    if (!m_datenempfangAngemeldet)
    {
      return;
    }

    AttributeGroup atgDaten;
    Aspect aspDaten;
    DataDescription ddDaten;

    // Anzeige vom Datenempfang anmelden

    atgDaten = m_dataModel.getAttributeGroup("atg.anzeigeEigenschaftSoll");
    aspDaten = m_dataModel.getAspect("asp.aqVorgabe");
    ddDaten = new DataDescription(atgDaten, aspDaten);

    m_debug.config("Abmeldung " + m_anzeige.getPid() + " " + ddDaten);

    m_connection.unsubscribeReceiver(empfaenger, m_anzeige, ddDaten);
  }

  /* (non-Javadoc)
   * @see stauma.dav.clientside.ClientReceiverInterface#update(stauma.dav.clientside.ResultData[])
   */
  public void update(ResultData[] results)
  {
    for (ResultData result : results)
    {
//      System.out.println("Anzeige Result " + result);

      anzeigeVerarbeiten(result, new Long(0));
    }
  }

  /**
   * Verarbeiten der Befehle für die Anzeige.
   *
   * @param result Befehle für die Anzeige.
   * @param zeitpunktUebergeordnet übergeordneter Schaltzeitpunkt.
   * @return Befehle für das Teilmodell KExTls.
   */
  public ResultData anzeigeVerarbeiten(ResultData result, Long zeitpunktUebergeordnet)
  {
    ResultData ergebnis = null;

//    System.out.println("Anzeige Result " + result);

    boolean datenVerarbeitet = false;

    SystemObject object = result.getObject();

    AttributeGroup atg = result.getDataDescription().getAttributeGroup();
    Aspect         asp = result.getDataDescription().getAspect();

    AnzeigeAktorTls anzeigeAktorUms = (AnzeigeAktorTls)m_listeAnzeigeAktorTlsKeyAnzeige.get(object.getPid());

    if (anzeigeAktorUms != null)
    {
      m_debug.fine("Umsetzung " + anzeigeAktorUms.getAnzeigeTypUmsetzung().getPid() + " " + result.toString());
//      m_debug.error("Umsetzung " + anzeigeAktorUms.getAnzeigeTypUmsetzung().getPid() + " " + result.toString());

      AnzeigeTypUmsetzungTls anzeigeTypUmsetzungTls = anzeigeAktorUms.getAnzeigeTypUmsetzung();

      int anzahlGlobalUmsetzung = anzeigeTypUmsetzungTls.getListeGlobalUmsetzung().size();

      for (int i = 0; i < anzahlGlobalUmsetzung; i++)
      {
        GlobalUmsetzung globalUmsetzung = (GlobalUmsetzung)anzeigeTypUmsetzungTls.getListeGlobalUmsetzung().get(i);

        if (globalUmsetzung.getUmsetzungsModul() != null)
        {
          ergebnis = umsetzungAbarbeiten(globalUmsetzung.getUmsetzungsModul(), result, zeitpunktUebergeordnet, object, atg, asp, anzeigeTypUmsetzungTls, i);

          if (ergebnis != null)
          {
            datenVerarbeitet = true;

            break;
          }
        }
      }

      if (!datenVerarbeitet)
      {
        ergebnis = umsetzungAbarbeiten(klassenNameDefault, result, zeitpunktUebergeordnet, object, atg, asp, anzeigeTypUmsetzungTls, -1);
      }
    }

    return ergebnis;
  }

  /* (non-Javadoc)
   * @see stauma.dav.clientside.ClientSenderInterface#dataRequest(stauma.dav.configuration.interfaces.SystemObject, stauma.dav.clientside.DataDescription, byte)
   */
  public void dataRequest(SystemObject object, DataDescription dataDescription, byte state)
  {
  }

  /* (non-Javadoc)
   * @see stauma.dav.clientside.ClientSenderInterface#isRequestSupported(stauma.dav.configuration.interfaces.SystemObject, stauma.dav.clientside.DataDescription)
   */
  public boolean isRequestSupported(SystemObject object, DataDescription dataDescription)
  {
    return false;
  }

  /**
   * Methode für die Abarbeitung der Umsetzung für die Anzeige.
   *
   * @param klassenName Name der Klasse für die Umsetzung.
   * @param result Eingangsdaten.
   * @param zeitpunktUebergeordnet übergeordneter Schaltzeitpunkt.
   * @param object Anzeige
   * @param atg AttributGruppe der Eingangsdaten.
   * @param asp Aspekt der Eingangsdaten.
   * @param anzeigeTypUmsetzungTls AnzeigeTypUmsetzungTls für diese Anzeige.
   * @param indexUmsetzung Index der Umsetzung.
   * @return Befehle für das Teilmodell KExTls.
   */
  public ResultData umsetzungAbarbeiten(String klassenName, ResultData result, Long zeitpunktUebergeordnet, SystemObject object, AttributeGroup atg, Aspect asp, AnzeigeTypUmsetzungTls anzeigeTypUmsetzungTls, int indexUmsetzung)
  {
    ResultData ergebnis = null;
    Class klasse = null;
    String aktuelleMethode = null;

    m_debug.fine("anzeigeUmsetzungAbarbeiten " + klassenName + " object " + object + " result " + result);

    try
    {
      klasse = Class.forName(klassenName);

//      System.out.println("klasse " + klasse);

      Class[] formaleParameter = new Class[4];
      formaleParameter[0] = HashMap.class;
      formaleParameter[1] = HashMap.class;
      formaleParameter[2] = boolean.class;
      formaleParameter[3] = boolean.class;

      Constructor construktor = klasse.getConstructor(formaleParameter);
      Object ret = construktor.newInstance(new Object[]{m_listeAnzeigeAktorTlsKeyDes, m_listeAnzeigeAktorTlsKeyAnzeige, isKannFreieGrafik(), isKannFreiText()});

      Method methode2 = null;

      for (int i = 0; i < methoden.length; i++)
      {
        if (atg.getPid().equals(methoden[i][0]))
        {
          aktuelleMethode = methoden[i][1];

//          System.out.println("methode " + aktuelleMethode);

          methode2 = klasse.getMethod(aktuelleMethode, new Class[]{ResultData.class, Long.class});

          break;
        }
      }

      if (methode2 != null)
      {
        ergebnis = (ResultData)methode2.invoke(ret, new Object[]{result, zeitpunktUebergeordnet});
      }
    }
    catch (ClassNotFoundException e)
    {
      m_debug.error("Klasse '" + klasse + "' für das Abarbeiten der Anzeigenumsetzung konnte nicht geladen werden.");

      e.printStackTrace();
    }
    catch (NoSuchMethodException e)
    {
      m_debug.error("Methode '" + aktuelleMethode + "()' in der Klasse '" + klasse + "' existiert nicht.");

      e.printStackTrace();
    }
    catch (InvocationTargetException e)
    {
      m_debug.error("InvocationTargetException");

      e.printStackTrace();
    }
    catch (IllegalAccessException e)
    {
      m_debug.error("Fehler beim Zugriff auf die Methode '" + aktuelleMethode + "()' in der Klasse '" + klasse + "'.");

      e.printStackTrace();
    }
    catch (InstantiationException e)
    {
      m_debug.error("Fehler bei der Instanziierung der Methode '" + aktuelleMethode + "()' in der Klasse '" + klasse + "'.");

      e.printStackTrace();
    }

    return ergebnis;
  }

  /**
   * Holen der Information, ob die Anzeige freie Grafik darstellen kann.
   *
   * @return <code>true</code>, wenn die Anzeige freie Grafik darstellen kann, sonst <code>false</code>.
   */
  public boolean isKannFreieGrafik()
  {
    return m_kannFreieGrafik;
  }

  /**
   * Setzen der Information, ob die Anzeige freie Grafik darstellen kann.
   *
   * @param kannFreieGrafik <code>true</code>, wenn die Anzeige freie Grafik darstellen kann, sonst <code>false</code>.
   */
  public void setKannFreieGrafik(boolean kannFreieGrafik)
  {
    m_kannFreieGrafik = kannFreieGrafik;
  }

  /**
   * Holen der Information, ob die Anzeige freien Text darstellen kann.
   *
   * @return <code>true</code>, wenn die Anzeige freien Text darstellen kann, sonst <code>false</code>.
   */
  public boolean isKannFreiText()
  {
    return m_kannFreiText;
  }

  /**
   * Setzen der Information, ob die Anzeige freien Text darstellen kann.
   *
   * @param kannFreiText <code>true</code>, wenn die Anzeige freien Text darstellen kann, sonst <code>false</code>.
   */
  public void setKannFreiText(boolean kannFreiText)
  {
    m_kannFreiText = kannFreiText;
  }

  /**
   * Holen des AnzeigeQuerschnitt, dem die Anzeige angehört.
   *
   * @return AnzeigeQuerschnitt, dem die Anzeige angehört.
   */
  public AnzeigeQuerschnitt getAnzeigeQuerschnitt()
  {
    return m_anzeigeQuerschnitt;
  }

  /**
   * Setzen des AnzeigeQuerschnitt, dem die Anzeige angehört.
   *
   * @param anzeigeQuerschnitt AnzeigeQuerschnitt, dem die Anzeige angehört.
   */
  public void setAnzeigeQuerschnitt(AnzeigeQuerschnitt anzeigeQuerschnitt)
  {
    m_anzeigeQuerschnitt = anzeigeQuerschnitt;
  }

  /**
   * Holen des Empfängers für die Befehle für die Anzeige.
   *
   * @return Empfänger für die Befehle für die Anzeige.
   */
  public ClientReceiverInterface getEmpfaenger()
  {
    ClientReceiverInterface empfaenger = this;

    if (getAnzeigeQuerschnitt() != null)
    {
      empfaenger = getAnzeigeQuerschnitt().getEmpfaenger();
    }

    return empfaenger;
  }

  /**
   * Holen des Senders für die Befehle für die DE.
   *
   * @return Senders für die Befehle für die DE.
   */
  public ClientSenderInterface getSender()
  {
    ClientSenderInterface sender = this;

    return sender;
  }

  /**
   * Holen der Pid der Anzeige.
   *
   * @return Pid der Anzeige.
   */
  public String getPid()
  {
    return m_anzeige.getPid();
  }

  /**
   * Anmelden zum Datenempfang für den Empfänger.
   */
  public void anmeldungDatenempfang()
  {
    anmeldungDatenempfang(getEmpfaenger());
  }
  /**
   * Anmelden zum Datenversand für den Sender.
   */
  public void anmeldungDatenversand()
  {
    anmeldungDatenversand(getSender());
  }

  /**
   * Senden von Daten über den Sender.
   *
   * @param resultData zu versendende Daten.
   * @param zeitpunkt übergeordneter Schaltzeitpunkt.
   * @return <code>true</code>.
   */
  public boolean datenSenden(ResultData resultData, long zeitpunkt)
  {
    if (zeitpunkt != 0)
    {
      resultData.setDataTime(zeitpunkt);
    }
    
    try
    {
      m_connection.sendData(resultData);
    }
    catch (DataNotSubscribedException e)
    {
      e.printStackTrace();
    }
    catch (SendSubscriptionNotConfirmed e)
    {
      e.printStackTrace();
    }
    
    return true;
  }

  /**
   * Holen der HashMap mit den AnzeigeAktorTls mit Key Anzeige.
   * 
   * @return HashMap mit den AnzeigeAktorTls mit Key Anzeige.
   */
  public HashMap<String, AnzeigeAktorTls> getListeAnzeigeAktorTlsKeyAnzeige()
  {
    return m_listeAnzeigeAktorTlsKeyAnzeige;
  }
}
